@@ -1,6 +1,5 @@
 import json
 import logging
-import os
 import re
 from typing import (
     Any,
@@ -13,20 +12,22 @@ from typing import (
     Set,
     TypedDict,
     Union,
-    cast,
 )
 from uuid import UUID
 
 import braintrust
 from braintrust import NOOP_SPAN, Logger, Span, SpanAttributes, SpanTypeAttribute, current_span, init_logger
+from braintrust.version import VERSION as sdk_version
 from langchain_core.agents import AgentAction, AgentFinish
 from langchain_core.callbacks.base import BaseCallbackHandler
 from langchain_core.documents import Document
-from langchain_core.messages import BaseMessage, ToolMessage
+from langchain_core.messages import BaseMessage
 from langchain_core.outputs.llm_result import LLMResult
 from tenacity import RetryCallState
 from typing_extensions import NotRequired
 
+from braintrust_langchain.version import version
+
 _logger = logging.getLogger("braintrust_langchain")
 
 
@@ -54,7 +55,7 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
     ):
         self.logger = logger
         self.spans: Dict[UUID, Span] = {}
-        self.debug = bool(os.environ.get("DEBUG")) or debug
+        self.debug = debug  # DEPRECATED
         self.exclude_metadata_props = exclude_metadata_props or re.compile(
             r"^(l[sc]_|langgraph_|__pregel_|checkpoint_ns)"
         )
@@ -104,7 +105,14 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
             "metadata": {
                 **({"tags": tags}),
                 **(event.get("metadata") or {}),
-                **({"runId": run_id, "parentRunId": parent_run_id} if self.debug else {}),
+                "run_id": run_id,
+                "parent_run_id": parent_run_id,
+                "braintrust": {
+                    "integration_name": "langchain-py",
+                    "integration_version": version,
+                    "sdk_version": sdk_version,
+                    "language": "python",
+                },
             },
         }
 
@@ -137,6 +145,7 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
         self.spans[run_id] = span
         return span
 
+    # TODO: serialize input, output, metadata correctly
     def _end_span(
         self,
         run_id: UUID,
@@ -180,9 +189,6 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
 
         span.end()
 
-    def clean_metadata(self, metadata: Optional[Mapping[str, Any]]) -> Mapping[str, Any]:
-        return {k: v for k, v in (metadata or {}).items() if not self.exclude_metadata_props.search(k)}
-
     def on_llm_error(
         self,
         error: BaseException,
@@ -191,7 +197,7 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
         parent_run_id: Optional[UUID] = None,
         **kwargs: Any,  # TODO: response=
     ) -> Any:
-        self._end_span(run_id, error=str(error))
+        self._end_span(run_id, error=str(error), metadata={**kwargs})
 
     def on_chain_error(
         self,
@@ -201,7 +207,7 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
         parent_run_id: Optional[UUID] = None,
         **kwargs: Any,  # TODO: some metadata
     ) -> Any:
-        self._end_span(run_id, error=str(error))
+        self._end_span(run_id, error=str(error), metadata={**kwargs})
 
     def on_tool_error(
         self,
@@ -211,7 +217,7 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
         parent_run_id: Optional[UUID] = None,
         **kwargs: Any,
     ) -> Any:
-        self._end_span(run_id, error=str(error))
+        self._end_span(run_id, error=str(error), metadata={**kwargs})
 
     def on_retriever_error(
         self,
@@ -221,7 +227,7 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
         parent_run_id: Optional[UUID] = None,
         **kwargs: Any,
     ) -> Any:
-        self._end_span(run_id, error=str(error))
+        self._end_span(run_id, error=str(error), metadata={**kwargs})
 
     # Agent Methods
     def on_agent_action(
@@ -237,9 +243,7 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
             run_id,
             type=SpanTypeAttribute.LLM,
             name=action.tool,
-            event={
-                "input": action.tool_input,  # type: ignore[arg-type]
-            },
+            event={"input": action, "metadata": {**kwargs}},
         )
 
     def on_agent_finish(
@@ -250,7 +254,7 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
         parent_run_id: Optional[UUID] = None,
         **kwargs: Any,
     ) -> Any:
-        self._end_span(run_id, output=finish.return_values)  # type: ignore
+        self._end_span(run_id, output=finish, metadata={**kwargs})
 
     def on_chain_start(
         self,
@@ -264,6 +268,13 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
         metadata: Optional[Dict[str, Any]] = None,
         **kwargs: Any,
     ) -> Any:
+        tags = tags or []
+
+        # avoids extra logs that seem not as useful esp. with langgraph
+        if "langsmith:hidden" in tags:
+            self.skipped_runs.add(run_id)
+            return
+
         metadata = metadata or {}
         resolved_name = (
             metadata.get("langgraph_node")
@@ -273,18 +284,20 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
             or "Chain"
         )
 
-        tags = tags or []
-
-        # avoids extra logs that seem not as useful esp. with langgraph
-        if "langsmith:hidden" in tags:
-            self.skipped_runs.add(run_id)
-            return
-
         self._start_span(
             parent_run_id,
             run_id,
             name=resolved_name,
-            event={"input": inputs, "metadata": {"tags": tags, **metadata, **kwargs}},
+            event={
+                "input": inputs,
+                "tags": tags,
+                "metadata": {
+                    "serialized": serialized,
+                    "name": name,
+                    "metadata": metadata,
+                    **kwargs,
+                },
+            },
         )
 
     def on_chain_end(
@@ -296,7 +309,7 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
         tags: Optional[List[str]] = None,
         **kwargs: Any,
     ) -> Any:
-        self._end_span(run_id, output=output_from_chain_values(outputs), tags=tags)
+        self._end_span(run_id, output=outputs, tags=tags, metadata={**kwargs})
 
     def on_llm_start(
         self,
@@ -321,8 +334,10 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
                 "input": prompts,
                 "tags": tags,
                 "metadata": {
-                    **self.clean_metadata(metadata),
-                    **kwargs["invocation_params"],
+                    "serialized": serialized,
+                    "name": name,
+                    "metadata": metadata,
+                    **kwargs,
                 },
             },
         )
@@ -348,13 +363,15 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
             name=name or serialized.get("name") or last_item(serialized.get("id") or []) or "Chat Model",
             type=SpanTypeAttribute.LLM,
             event={
-                "input": input_from_messages(messages),
+                "input": messages,
                 "tags": tags,
-                "metadata": clean_object(
+                "metadata": (
                     {
-                        **self.clean_metadata(metadata),
-                        **extract_call_args(serialized, invocation_params, metadata),
-                        "tools": invocation_params.get("tools"),
+                        "serialized": serialized,
+                        "invocation_params": invocation_params,
+                        "metadata": metadata or {},
+                        "name": name,
+                        **kwargs,
                     }
                 ),
             },
@@ -372,16 +389,18 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
         if run_id not in self.spans:
             return
 
-        metadata = {k: v for k, v in _to_dict(response).items() if k not in ("llm_output", "generations")}
         metrics = _get_metrics_from_response(response)
         model_name = _get_model_name_from_response(response)
 
         self._end_span(
             run_id,
-            output=output_from_generations(response.generations),
+            output=response,
             metrics=metrics,
             tags=tags,
-            metadata=self.clean_metadata({**metadata, "model": model_name}),
+            metadata={
+                "model": model_name,
+                **kwargs,
+            },
         )
 
     def on_tool_start(
@@ -405,8 +424,13 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
                 "input": inputs or safe_parse_serialized_json(input_str),
                 "tags": tags,
                 "metadata": {
-                    **self.clean_metadata(metadata),
-                    **extract_call_args(serialized, kwargs.get("invocation_params") or {}, metadata),
+                    "metadata": metadata,
+                    "serialized": serialized,
+                    "input_str": input_str,
+                    "input": safe_parse_serialized_json(input_str),
+                    "inputs": inputs,
+                    "name": name,
+                    **kwargs,
                 },
             },
         )
@@ -419,7 +443,7 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
         parent_run_id: Optional[UUID] = None,
         **kwargs: Any,
     ) -> Any:
-        self._end_span(run_id, output=output_from_tool_output(output))
+        self._end_span(run_id, output=output, metadata={**kwargs})
 
     def on_retriever_start(
         self,
@@ -442,8 +466,10 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
                 "input": query,
                 "tags": tags,
                 "metadata": {
-                    **self.clean_metadata(metadata),
-                    **extract_call_args(serialized, kwargs.get("invocation_params") or {}, metadata),
+                    "serialized": serialized,
+                    "metadata": metadata,
+                    "name": name,
+                    **kwargs,
                 },
             },
         )
@@ -456,7 +482,7 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
         parent_run_id: Optional[UUID] = None,
         **kwargs: Any,
     ) -> Any:
-        self._end_span(run_id, output=documents)
+        self._end_span(run_id, output=documents, metadata={**kwargs})
 
     def on_llm_new_token(
         self,
@@ -502,83 +528,6 @@ class BraintrustCallbackHandler(BaseCallbackHandler):
         pass
 
 
-def extract_call_args(
-    llm: Dict[str, Any],
-    invocation_params: Dict[str, Any],
-    metadata: Optional[Dict[str, Any]] = None,
-) -> Dict[str, Any]:
-    metadata = metadata or {}
-
-    # NOTE: These vary by langchain model used. We try to normalize them here.
-    args = clean_object(
-        {
-            "model": pick(invocation_params.get("model"), metadata.get("ls_model_name"), llm.get("name")),
-            "temperature": pick(invocation_params.get("temperature"), metadata.get("ls_temperature")),
-            "top_p": pick(invocation_params.get("top_p"), invocation_params.get("top_k")),
-            "top_k": pick(invocation_params.get("top_k"), invocation_params.get("top_p")),
-            "max_tokens": pick(invocation_params.get("max_tokens"), invocation_params.get("max_output_tokens")),
-            "frequency_penalty": invocation_params.get("frequency_penalty"),
-            "presence_penalty": invocation_params.get("presence_penalty"),
-            "response_format": invocation_params.get("response_format"),
-            "tool_choice": invocation_params.get("tool_choice"),
-            "function_call": invocation_params.get("function_call"),
-            "n": invocation_params.get("n"),
-            "stop": pick(invocation_params.get("stop"), invocation_params.get("stop_sequence")),
-        }
-    )
-
-    # Failsafe let's provide the invocation params as is
-    return invocation_params if not args else args
-
-
-def pick(*values: Any) -> Any:
-    return next((value for value in values if value is not None), None)
-
-
-def output_from_generations(generations: Union[List[List[Any]], List[Any]]) -> List[Any]:
-    parsed: List[Any] = []
-    for batch in generations:
-        if isinstance(batch, list):
-            parsed.extend(map(parse_generation, batch))  # pyright: ignore
-        else:
-            parsed.append(parse_generation(batch))
-    return parsed
-
-
-def parse_generation(generation: Any) -> Any:
-    if hasattr(generation, "message"):
-        return get_message_content(generation.message)
-    if hasattr(generation, "text"):
-        return generation.text
-    # Give up!
-    return None
-
-
-def input_from_messages(messages: List[List[Any]]) -> List[Any]:
-    return [get_message_content(message) for batch in messages for message in batch]
-
-
-def get_message_content(message: Any) -> Dict[str, Any]:
-    role = getattr(message, "name", None) or message.type
-
-    if message.type == "human":
-        role = "user"
-    elif message.type == "ai":
-        role = "assistant"
-    elif message.type == "system":
-        role = "system"
-
-    return clean_object(
-        {
-            "content": message.content,
-            "role": role,
-            "tool_calls": getattr(message, "tool_calls", None),
-            "status": getattr(message, "status", None),
-            "artifact": getattr(message, "artifact", None),
-        }
-    )
-
-
 def clean_object(obj: Dict[str, Any]) -> Dict[str, Any]:
     return {
         k: v
@@ -594,64 +543,6 @@ def safe_parse_serialized_json(input_str: str) -> Any:
         return input_str
 
 
-def output_from_tool_output(output: Any) -> Optional[Dict[str, Any]]:
-    return get_message_content(output) if isinstance(output, ToolMessage) else None
-
-
-def flatten_list(items: List[Any]) -> List[Any]:
-    result: List[Any] = []
-    for item in items:
-        if isinstance(item, list):
-            result.extend(cast(List[Any], item))
-        else:
-            result.append(item)
-    return result
-
-
-def output_from_chain_values(output: Any) -> Any:
-    output_list: List[Any] = [output] if not isinstance(output, list) else output
-
-    processed = [parse_chain_value(x) for x in output_list]
-
-    parsed = flatten_list(processed)
-
-    return parsed[0] if len(parsed) == 1 else parsed
-
-
-def parse_chain_value(output: Any) -> Any:
-    if isinstance(output, str):
-        return output
-
-    if not output:
-        return output
-
-    if hasattr(output, "content"):
-        return output.content
-
-    if hasattr(output, "messages"):
-        return [parse_chain_value(m) for m in output.messages]
-
-    if hasattr(output, "value"):
-        return output.value
-
-    if hasattr(output, "kwargs"):
-        return parse_chain_value(output.kwargs)
-
-    # XXX: RunnableMap returns an object with keys for each sequence
-    if isinstance(output, dict):
-        output = cast(Dict[str, Any], output)
-        return {k: parse_chain_value(v) for k, v in cast(Dict[str, Any], output).items()}
-
-    # Give up! Let's assume the user will use the raw output.
-    return output
-
-
-def input_from_chain_values(inputs: Any) -> Any:
-    inputs_list = [inputs] if not isinstance(inputs, list) else inputs
-    parsed = [parse_chain_value(x) for x in inputs_list]
-    return parsed[0] if len(parsed) == 1 else parsed
-
-
 def last_item(items: List[Any]) -> Any:
     return items[-1] if items else None
 
@@ -709,11 +600,3 @@ def _get_metrics_from_response(response: LLMResult):
         metrics = llm_output.get("token_usage") or llm_output.get("estimatedTokens") or {}
 
     return clean_object(metrics)
-
-
-def _to_dict(obj: Any) -> Dict[str, Any]:
-    """Convert a Pydantic object to dict, with backwards compatibility for v1/v2."""
-    if hasattr(obj, "model_dump"):
-        return obj.model_dump()  # Pydantic v2
-    else:
-        return obj.dict()  # Pydantic v1
